#!/usr/bin/env python
#
# File: Psi4GenerateConformers.py
# Author: Manish Sud <msud@san.rr.com>
#
# Copyright (C) 2024 Manish Sud. All rights reserved.
#
# The functionality available in this script is implemented using Psi4, an
# open source quantum chemistry software package, and RDKit, an open
# source toolkit for cheminformatics developed by Greg Landrum.
#
# This file is part of MayaChemTools.
#
# MayaChemTools is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# MayaChemTools is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability of fitness
# for a particular purpose.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MayaChemTools; if not, see <http://www.gnu.org/licenses/> or
# write to the Free Software Foundation Inc., 59 Temple Place, Suite 330,
# Boston, MA, 02111-1307, USA.
#

from __future__ import print_function

# Add local python path to the global path and import standard library modules...
import os
import sys;  sys.path.insert(0, os.path.join(os.path.dirname(sys.argv[0]), "..", "lib", "Python"))
import time
import re
import shutil
import multiprocessing as mp

# Psi4 imports...
if (hasattr(shutil, 'which') and shutil.which("psi4") is None):
    sys.stderr.write("\nWarning: Failed to find 'psi4' in your PATH indicating potential issues with your\n")
    sys.stderr.write("Psi4 environment. The 'import psi4' directive in the global scope of the script\n")
    sys.stderr.write("interferes with the multiprocessing functionality. It is imported later in the\n")
    sys.stderr.write("local scope during the execution of the script and may fail. Check/update your\n")
    sys.stderr.write("Psi4 environment and try again.\n\n")

# RDKit imports...
try:
    from rdkit import rdBase
    from rdkit import Chem
    from rdkit.Chem import AllChem
except ImportError as ErrMsg:
    sys.stderr.write("\nFailed to import RDKit module/package: %s\n" % ErrMsg)
    sys.stderr.write("Check/update your RDKit environment and try again.\n\n")
    sys.exit(1)

# MayaChemTools imports...
try:
    from docopt import docopt
    import MiscUtil
    import Psi4Util
    import RDKitUtil
except ImportError as ErrMsg:
    sys.stderr.write("\nFailed to import MayaChemTools module/package: %s\n" % ErrMsg)
    sys.stderr.write("Check/update your MayaChemTools environment and try again.\n\n")
    sys.exit(1)

ScriptName = os.path.basename(sys.argv[0])
Options = {}
OptionsInfo = {}

def main():
    """Start execution of the script."""
    
    MiscUtil.PrintInfo("\n%s (Psi4: Imported later; RDKit v%s; MayaChemTools v%s; %s): Starting...\n" % (ScriptName, rdBase.rdkitVersion, MiscUtil.GetMayaChemToolsVersion(), time.asctime()))
    
    (WallClockTime, ProcessorTime) = MiscUtil.GetWallClockAndProcessorTime()
    
    # Retrieve command line arguments and options...
    RetrieveOptions()
    
    # Process and validate command line arguments and options...
    ProcessOptions()
    
    # Perform actions required by the script...
    GenerateConformers()
    
    MiscUtil.PrintInfo("\n%s: Done...\n" % ScriptName)
    MiscUtil.PrintInfo("Total time: %s" % MiscUtil.GetFormattedElapsedTime(WallClockTime, ProcessorTime))

def GenerateConformers():
    """Generate conformers."""
    
    # Setup a molecule reader...
    MiscUtil.PrintInfo("\nProcessing file %s..." % OptionsInfo["Infile"])
    Mols  = RDKitUtil.ReadMolecules(OptionsInfo["Infile"], **OptionsInfo["InfileParams"])
    
    # Set up a molecule writer...
    Writer = RDKitUtil.MoleculesWriter(OptionsInfo["Outfile"], **OptionsInfo["OutfileParams"])
    if Writer is None:
        MiscUtil.PrintError("Failed to setup a writer for output fie %s " % OptionsInfo["Outfile"])
    MiscUtil.PrintInfo("Generating file %s..." % OptionsInfo["Outfile"])

    MolCount, ValidMolCount, ConfGenFailedCount = ProcessMolecules(Mols, Writer)

    if Writer is not None:
        Writer.close()
    
    MiscUtil.PrintInfo("\nTotal number of molecules: %d" % MolCount)
    MiscUtil.PrintInfo("Number of valid molecules: %d" % ValidMolCount)
    MiscUtil.PrintInfo("Number of molecules failed during conformation generation or minimization: %d" % ConfGenFailedCount)
    MiscUtil.PrintInfo("Number of ignored molecules: %d" % (MolCount - ValidMolCount + ConfGenFailedCount))

def ProcessMolecules(Mols, Writer):
    """Process molecules to generate conformers."""
    
    if OptionsInfo["MPMode"]:
        return ProcessMoleculesUsingMultipleProcesses(Mols, Writer)
    else:
        return ProcessMoleculesUsingSingleProcess(Mols, Writer)

def ProcessMoleculesUsingSingleProcess(Mols, Writer):
    """Process molecules to generate conformers using a single process."""
    
    # Intialize Psi4...
    MiscUtil.PrintInfo("\nInitializing Psi4...")
    Psi4Handle = Psi4Util.InitializePsi4(Psi4RunParams = OptionsInfo["Psi4RunParams"], Psi4OptionsParams = OptionsInfo["Psi4OptionsParams"], PrintVersion = True, PrintHeader = True)
    OptionsInfo["psi4"] = Psi4Handle

    # Setup max iterations global variable...
    Psi4Util.UpdatePsi4OptionsParameters(Psi4Handle, {'GEOM_MAXITER': OptionsInfo["MaxIters"]})
    
    # Setup conversion factor for energy units...
    SetupEnergyConversionFactor(Psi4Handle)

    MiscUtil.PrintInfo("\nGenerating conformers and performing energy minimization...")

    (MolCount, ValidMolCount, ConfGenFailedCount) = [0] * 3
    for Mol in Mols:
        MolCount += 1

        if not CheckAndValidateMolecule(Mol, MolCount):
            continue

        # Setup 2D coordinates for SMILES input file...
        if OptionsInfo["SMILESInfileStatus"]:
            AllChem.Compute2DCoords(Mol)
        
        ValidMolCount += 1

        ConformerMol, CalcStatus, ConfIDs, ConfEnergies = GenerateMolConformers(Psi4Handle, Mol, MolCount)

        if not CalcStatus:
            if not OptionsInfo["QuietMode"]:
                MiscUtil.PrintWarning("Failed to calculate energy for molecule %s" % RDKitUtil.GetMolName(Mol, MolCount))
            
            ConfGenFailedCount += 1
            continue

        WriteMolConformers(Writer, ConformerMol, MolCount, ConfIDs, ConfEnergies)
        
    return (MolCount, ValidMolCount, ConfGenFailedCount)

def ProcessMoleculesUsingMultipleProcesses(Mols, Writer):
    """Process and minimize molecules using multiprocessing."""
    
    if OptionsInfo["MPLevelConformersMode"]:
        return ProcessMoleculesUsingMultipleProcessesAtConformersLevel(Mols, Writer)
    elif OptionsInfo["MPLevelMoleculesMode"]:
        return ProcessMoleculesUsingMultipleProcessesAtMoleculesLevel(Mols, Writer)
    else:
        MiscUtil.PrintError("The value, %s,  option \"--mpLevel\" is not supported." % (OptionsInfo["MPLevel"]))
        
def ProcessMoleculesUsingMultipleProcessesAtMoleculesLevel(Mols, Writer):
    """Process molecules to generate conformers using multiprocessing at molecules level."""

    MiscUtil.PrintInfo("\nGenerating conformers and performing energy minimization using multiprocessing at molecules level...")

    MPParams = OptionsInfo["MPParams"]
    
    # Setup data for initializing a worker process...
    InitializeWorkerProcessArgs = (MiscUtil.ObjectToBase64EncodedString(Options), MiscUtil.ObjectToBase64EncodedString(OptionsInfo))
    
    # Setup a encoded mols data iterable for a worker process...
    WorkerProcessDataIterable = RDKitUtil.GenerateBase64EncodedMolStrings(Mols)
    
    # Setup process pool along with data initialization for each process...
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("\nConfiguring multiprocessing using %s method..." % ("mp.Pool.imap()" if re.match("^Lazy$", MPParams["InputDataMode"], re.I) else "mp.Pool.map()"))
        MiscUtil.PrintInfo("NumProcesses: %s; InputDataMode: %s; ChunkSize: %s\n" % (MPParams["NumProcesses"], MPParams["InputDataMode"], ("automatic" if MPParams["ChunkSize"] is None else MPParams["ChunkSize"])))
    
    ProcessPool = mp.Pool(MPParams["NumProcesses"], InitializeWorkerProcess, InitializeWorkerProcessArgs)
    
    # Start processing...
    if re.match("^Lazy$", MPParams["InputDataMode"], re.I):
        Results = ProcessPool.imap(WorkerProcess, WorkerProcessDataIterable, MPParams["ChunkSize"])
    elif re.match("^InMemory$", MPParams["InputDataMode"], re.I):
        Results = ProcessPool.map(WorkerProcess, WorkerProcessDataIterable, MPParams["ChunkSize"])
    else:
        MiscUtil.PrintError("The value, %s, specified for \"--inputDataMode\" is not supported." % (MPParams["InputDataMode"]))
    
    # Print out Psi4 version in the main process...
    MiscUtil.PrintInfo("\nInitializing Psi4...\n")
    Psi4Handle  = Psi4Util.InitializePsi4(PrintVersion = True, PrintHeader = False)
    OptionsInfo["psi4"] = Psi4Handle
    
    (MolCount, ValidMolCount, ConfGenFailedCount) = [0] * 3
    for Result in Results:
        MolCount += 1
        MolIndex, EncodedMol, CalcStatus, ConfIDs, ConfEnergies = Result
        
        if EncodedMol is None:
            continue
        ValidMolCount += 1

        if not CalcStatus:
            ConfGenFailedCount += 1
            continue
            
        Mol = RDKitUtil.MolFromBase64EncodedMolString(EncodedMol)
        WriteMolConformers(Writer, Mol, MolCount, ConfIDs, ConfEnergies)
        
    return (MolCount, ValidMolCount, ConfGenFailedCount)

def InitializeWorkerProcess(*EncodedArgs):
    """Initialize data for a worker process."""
    
    global Options, OptionsInfo

    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("Starting process (PID: %s)..." % os.getpid())
    
    # Decode Options and OptionInfo...
    Options = MiscUtil.ObjectFromBase64EncodedString(EncodedArgs[0])
    OptionsInfo = MiscUtil.ObjectFromBase64EncodedString(EncodedArgs[1])
    
    # Psi4 is initialized in the worker process to avoid creation of redundant Psi4
    # output files for each process...
    OptionsInfo["Psi4Initialized"]  = False

def WorkerProcess(EncodedMolInfo):
    """Process data for a worker process."""

    if not OptionsInfo["Psi4Initialized"]:
        InitializePsi4ForWorkerProcess()
    
    MolIndex, EncodedMol = EncodedMolInfo
    MolNum = MolIndex + 1
    
    CalcStatus = False
    ConfIDs = None
    ConfEnergies = None
    
    if EncodedMol is None:
        return [MolIndex, None, CalcStatus, ConfIDs, ConfEnergies]

    Mol = RDKitUtil.MolFromBase64EncodedMolString(EncodedMol)
    
    if not CheckAndValidateMolecule(Mol, MolNum):
        return [MolIndex, None, CalcStatus, ConfIDs, ConfEnergies]
    
    # Setup 2D coordinates for SMILES input file...
    if OptionsInfo["SMILESInfileStatus"]:
        AllChem.Compute2DCoords(Mol)
    
    Mol, CalcStatus, ConfIDs, ConfEnergies = GenerateMolConformers(OptionsInfo["psi4"], Mol, MolNum)

    return [MolIndex, RDKitUtil.MolToBase64EncodedMolString(Mol, PropertyPickleFlags = Chem.PropertyPickleOptions.MolProps | Chem.PropertyPickleOptions.PrivateProps), CalcStatus, ConfIDs, ConfEnergies]

def ProcessMoleculesUsingMultipleProcessesAtConformersLevel(Mols, Writer):
    """Process molecules to generate conformers using multiprocessing at conformers level."""

    MiscUtil.PrintInfo("\nPerforming minimization with generation of conformers using multiprocessing at conformers level...")

    (MolCount, ValidMolCount, ConfGenFailedCount) = [0] * 3
    for Mol in Mols:
        MolCount += 1
        
        if not CheckAndValidateMolecule(Mol, MolCount):
            continue

        # Setup 2D coordinates for SMILES input file...
        if OptionsInfo["SMILESInfileStatus"]:
            AllChem.Compute2DCoords(Mol)

        ValidMolCount += 1
        
        Mol, CalcStatus, ConfIDs, ConfEnergies = ProcessConformersUsingMultipleProcesses(Mol, MolCount)

        if not CalcStatus:
            ConfGenFailedCount += 1
            continue
        
        WriteMolConformers(Writer, Mol, MolCount, ConfIDs, ConfEnergies)
        
    return (MolCount, ValidMolCount, ConfGenFailedCount)

def ProcessConformersUsingMultipleProcesses(Mol, MolNum):
    """Generate coformers and minimize them using multiple processes."""

    # Add hydrogens...
    Mol = Chem.AddHs(Mol)

    # Setup conformers...
    ConfIDs = EmbedMolecule(Mol, MolNum)
    if not len(ConfIDs):
        if not OptionsInfo["QuietMode"]:
            MolName = RDKitUtil.GetMolName(Mol, MolNum)
            MiscUtil.PrintWarning("Minimization couldn't be performed for molecule %s: Embedding failed...\n" % MolName)
        return (Mol, False, None, None)

    MPParams = OptionsInfo["MPParams"]
    
    # Setup data for initializing a worker process...
    InitializeWorkerProcessArgs = (MiscUtil.ObjectToBase64EncodedString(Options), MiscUtil.ObjectToBase64EncodedString(OptionsInfo))

    # Setup a encoded mols data iterable for a worker process...
    MolIndex = MolNum - 1
    WorkerProcessDataIterable = RDKitUtil.GenerateBase64EncodedMolStringWithConfIDs(Mol, MolIndex, ConfIDs)

    # Setup process pool along with data initialization for each process...
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("\nConfiguring multiprocessing using %s method..." % ("mp.Pool.imap()" if re.match("^Lazy$", MPParams["InputDataMode"], re.I) else "mp.Pool.map()"))
        MiscUtil.PrintInfo("NumProcesses: %s; InputDataMode: %s; ChunkSize: %s\n" % (MPParams["NumProcesses"], MPParams["InputDataMode"], ("automatic" if MPParams["ChunkSize"] is None else MPParams["ChunkSize"])))
    
    ProcessPool = mp.Pool(MPParams["NumProcesses"], InitializeConformerWorkerProcess, InitializeWorkerProcessArgs)
    
    # Start processing...
    if re.match("^Lazy$", MPParams["InputDataMode"], re.I):
        Results = ProcessPool.imap(ConformerWorkerProcess, WorkerProcessDataIterable, MPParams["ChunkSize"])
    elif re.match("^InMemory$", MPParams["InputDataMode"], re.I):
        Results = ProcessPool.map(ConformerWorkerProcess, WorkerProcessDataIterable, MPParams["ChunkSize"])
    else:
        MiscUtil.PrintError("The value, %s, specified for \"--inputDataMode\" is not supported." % (MPParams["InputDataMode"]))
    
    CalcEnergyMap = {}
    CalcFailedCount = 0
    for Result in Results:
        MolIndex, EncodedMol, CalcStatus, ConfID, Energy = Result

        if EncodedMol is None:
            CalcFailedCount += 1
            continue
        
        if not CalcStatus:
            CalcFailedCount += 1
            continue
        
        # Retrieve minimized atom positions...
        MinimizedMol = RDKitUtil.MolFromBase64EncodedMolString(EncodedMol)
        AtomPositions = RDKitUtil.GetAtomPositions(MinimizedMol, ConfID = ConfID)

        # Update atom positions...
        RDKitUtil.SetAtomPositions(Mol, AtomPositions, ConfID = ConfID)
        
        CalcEnergyMap[ConfID] = Energy
    
    if CalcFailedCount:
        return (Mol, False, None, None)
    
    # Align molecules after minimization...
    if OptionsInfo["ConfGenerationParams"]["AlignConformers"]:
        AllChem.AlignMolConformers(Mol)

    # Filter conformers...
    SelectedConfIDs, SelectedConfEnergies = FilterMolConformers(Mol, MolNum, ConfIDs, CalcEnergyMap)
    
    return [Mol, True, SelectedConfIDs, SelectedConfEnergies]
    
def InitializeConformerWorkerProcess(*EncodedArgs):
    """Initialize data for a conformer worker process."""
    
    global Options, OptionsInfo
    
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("Starting process (PID: %s)..." % os.getpid())
    
    # Decode Options and OptionInfo...
    Options = MiscUtil.ObjectFromBase64EncodedString(EncodedArgs[0])
    OptionsInfo = MiscUtil.ObjectFromBase64EncodedString(EncodedArgs[1])
    
    # Psi4 is initialized in the worker process to avoid creation of redundant Psi4
    # output files for each process...
    OptionsInfo["Psi4Initialized"]  = False

def ConformerWorkerProcess(EncodedMolInfo):
    """Process data for a conformer worker process."""

    if not OptionsInfo["Psi4Initialized"]:
        InitializePsi4ForWorkerProcess()
    
    MolIndex, EncodedMol, ConfID = EncodedMolInfo

    MolNum = MolIndex + 1
    
    CalcStatus = False
    Energy = None
    
    if EncodedMol is None:
        return [MolIndex, None, CalcStatus, ConfID, Energy]
    
    Mol = RDKitUtil.MolFromBase64EncodedMolString(EncodedMol)
    MolName = RDKitUtil.GetMolName(Mol, MolNum)
    
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("Processing conformer ID %s for molecule %s..." % (ConfID, MolName))
    
    Status, ConvergeStatus = MinimizeMoleculeUsingForceField(Mol, MolNum, ConfID)
    if not Status:
        return [MolIndex, EncodedMol, CalcStatus, ConfID, Energy]
    
    if ConvergeStatus != 0:
        if not OptionsInfo["QuietMode"]:
            MolName = RDKitUtil.GetMolName(Mol, (MolIndex + 1))
            MiscUtil.PrintWarning("Minimization using forcefield failed to converge for molecule %s in %d steps. Try using higher value for \"maxIters\" in \"--confParams\" option...\n" % (MolName, OptionsInfo["ConfGenerationParams"]["MaxIters"]))
    
    # Perform Psi4 minimization...
    CalcStatus, Energy = MinimizeMoleculeUsingPsi4(OptionsInfo["psi4"], Mol, MolNum, ConfID)
    if not CalcStatus:
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintWarning("Minimization couldn't be performed for molecule %s\n" % (MolName))
            return [MolIndex, EncodedMol, False, ConfID, Energy]

    return [MolIndex, RDKitUtil.MolToBase64EncodedMolString(Mol, PropertyPickleFlags = Chem.PropertyPickleOptions.MolProps | Chem.PropertyPickleOptions.PrivateProps), CalcStatus, ConfID, Energy]

def InitializePsi4ForWorkerProcess():
    """Initialize Psi4 for a worker process."""
    
    if OptionsInfo["Psi4Initialized"]:
        return

    OptionsInfo["Psi4Initialized"] = True

    if OptionsInfo["MPLevelConformersMode"] and re.match("auto", OptionsInfo["Psi4RunParams"]["OutputFileSpecified"], re.I):
        # Run Psi4 in quiet mode during multiprocessing at Conformers level for 'auto' OutputFile...
        OptionsInfo["Psi4RunParams"]["OutputFile"] = "quiet"
    else:
        # Update output file...
        OptionsInfo["Psi4RunParams"]["OutputFile"] = Psi4Util.UpdatePsi4OutputFileUsingPID(OptionsInfo["Psi4RunParams"]["OutputFile"], os.getpid())
            
    # Intialize Psi4...
    OptionsInfo["psi4"] = Psi4Util.InitializePsi4(Psi4RunParams = OptionsInfo["Psi4RunParams"], Psi4OptionsParams = OptionsInfo["Psi4OptionsParams"], PrintVersion = False, PrintHeader = True)
    
    # Setup max iterations global variable...
    Psi4Util.UpdatePsi4OptionsParameters(OptionsInfo["psi4"], {'GEOM_MAXITER': OptionsInfo["MaxIters"]})
    
    # Setup conversion factor for energy units...
    SetupEnergyConversionFactor(OptionsInfo["psi4"])

def GenerateMolConformers(Psi4Handle, Mol, MolNum):
    """Generate and mininize conformers for a molecule."""

    # Add hydrogens..
    Mol = Chem.AddHs(Mol)
    
    MolName = RDKitUtil.GetMolName(Mol, MolNum)

    # Setup conformers...
    ConfIDs = EmbedMolecule(Mol, MolNum)
    if not len(ConfIDs):
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintWarning("Conformation generation couldn't be performed for molecule %s: Embedding failed...\n" % MolName)
        return (Mol, False, None, None)
    
    # Minimize conformers...
    CalcEnergyMap = {}
    for ConfID in ConfIDs:
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintInfo("Processing conformer ID %s for molecule %s..." % (ConfID, MolName))
            
        # Perform forcefield minimization...
        Status, ConvergeStatus = MinimizeMoleculeUsingForceField(Mol, MolNum, ConfID)
        if not Status:
            return (Mol, False, None, None)
        
        if ConvergeStatus != 0:
            if not OptionsInfo["QuietMode"]:
                MiscUtil.PrintWarning("Minimization using forcefield failed to converge for molecule %s in %d steps. Try using higher value for \"maxIters\" in \"--confParams\" option...\n" % (MolName, OptionsInfo["ConfGenerationParams"]["MaxIters"]))

        # Perform Psi4 minimization...
        CalcStatus, Energy = MinimizeMoleculeUsingPsi4(Psi4Handle, Mol, MolNum, ConfID)
        if not CalcStatus:
            if not OptionsInfo["QuietMode"]:
                MiscUtil.PrintWarning("Minimization couldn't be performed for molecule %s\n" % (MolName))
            return (Mol, False, None, None)

        CalcEnergyMap[ConfID] = Energy
    
    # Align molecules after minimization...
    if OptionsInfo["ConfGenerationParams"]["AlignConformers"]:
        AllChem.AlignMolConformers(Mol)

    # Filter conformers...
    SelectedConfIDs, SelectedConfEnergies = FilterMolConformers(Mol, MolNum, ConfIDs, CalcEnergyMap)
    
    return [Mol, True, SelectedConfIDs, SelectedConfEnergies]

def FilterMolConformers(Mol, MolNum, ConfIDs, CalcEnergyMap):
    """Filter conformers for a molecule."""

    SortedConfIDs = sorted(ConfIDs, key = lambda ConfID: CalcEnergyMap[ConfID])

    MinEnergyConfID = SortedConfIDs[0]
    MinConfEnergy = CalcEnergyMap[MinEnergyConfID]
    EnergyWindow = OptionsInfo["EnergyWindow"]
    
    EnergyRMSDCutoff = OptionsInfo["EnergyRMSDCutoff"]
    ApplyEnergyRMSDCutoff = False
    if EnergyRMSDCutoff > 0:
        ApplyEnergyRMSDCutoff = True

    EnergyRMSDCutoffLowest = OptionsInfo["EnergyRMSDCutoffModeLowest"]
    EnergyRMSDCalcModeBest = OptionsInfo["EnergyRMSDCalcModeBest"]
    
    PreAligned = False
    if OptionsInfo["ConfGenerationParams"]["AlignConformers"]:
        PreAligned = True
    
    RefMol, ProbeMol = [None] * 2
    if EnergyRMSDCalcModeBest:
        # Copy molecules for best RMSD calculations to avoid change in the coordinates
        # of the conformations...
        RefMol = AllChem.Mol(Mol)
        ProbeMol = AllChem.Mol(Mol)
    
    # Track conformers with in the specified energy window  from the lowest
    # energy conformation along with applying RMSD cutoff as needed...
    #
    SelectedConfIDs = []
    
    ConfCount = 0
    IgnoredByEnergyConfCount = 0
    IgnoredByRMSDConfCount = 0
    
    FirstConf = True
    
    for ConfID in SortedConfIDs:
        if FirstConf:
            FirstConf = False
            ConfCount += 1
            SelectedConfIDs.append(ConfID)
            continue
            
        ConfEnergyDiff = abs(CalcEnergyMap[ConfID] - MinConfEnergy )
        if  ConfEnergyDiff > EnergyWindow:
            IgnoredByEnergyConfCount += 1
            continue
        
        if ApplyEnergyRMSDCutoff:
            IgnoreConf = False
            if EnergyRMSDCutoffLowest:
                # Compare RMSD with the lowest energy conformation...
                if EnergyRMSDCalcModeBest:
                    CalcRMSD = AllChem.GetBestRMS(ProbeMol, RefMol, prbId = ConfID, refId = MinEnergyConfID)
                else:
                    CalcRMSD = AllChem.GetConformerRMS(Mol, MinEnergyConfID, ConfID, prealigned=PreAligned)
                if CalcRMSD < EnergyRMSDCutoff:
                        IgnoreConf = True
            else:
                for SelectedConfID in SelectedConfIDs:
                    if EnergyRMSDCalcModeBest:
                        CalcRMSD = AllChem.GetBestRMS(ProbeMol, RefMol, prbId = ConfID, refId = SelectedConfID)
                    else:
                        CalcRMSD = AllChem.GetConformerRMS(Mol, SelectedConfID, ConfID, prealigned=PreAligned)
                    if CalcRMSD < EnergyRMSDCutoff:
                        IgnoreConf = True
                        break
            if IgnoreConf:
                IgnoredByRMSDConfCount += 1
                continue
                
        ConfCount += 1
        SelectedConfIDs.append(ConfID)
    
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("\nTotal Number of conformations generated for %s: %d" % (RDKitUtil.GetMolName(Mol, MolNum), ConfCount))
        MiscUtil.PrintInfo("Number of conformations ignored due to energy window cutoff: %d" % (IgnoredByEnergyConfCount))
        if ApplyEnergyRMSDCutoff:
            MiscUtil.PrintInfo("Number of conformations ignored due to energy RMSD cutoff:  %d" % (IgnoredByRMSDConfCount))

    SelectedConfEnergies = None
    if OptionsInfo["EnergyOut"]:
        SelectedConfEnergies = []
        for ConfID in SelectedConfIDs:
            Energy = "%.*f" % (OptionsInfo["Precision"], CalcEnergyMap[ConfID])
            SelectedConfEnergies.append(Energy)
    
    return [SelectedConfIDs, SelectedConfEnergies]

def MinimizeMoleculeUsingPsi4(Psi4Handle, Mol, MolNum, ConfID = -1):
    """Minimize molecule using Psi4."""

    # Setup a Psi4Mol...
    Psi4Mol = SetupPsi4Mol(Psi4Handle, Mol, MolNum, ConfID)
    if Psi4Mol is None:
        return (False, None)
        
    #  Setup reference wave function...
    Reference = SetupReferenceWavefunction(Mol)
    Psi4Handle.set_options({'Reference': Reference})
    
    # Setup method name and basis set...
    MethodName, BasisSet = SetupMethodNameAndBasisSet(Mol)

    # Optimize geometry...
    Status, Energy, WaveFunction = Psi4Util.PerformGeometryOptimization(Psi4Handle, Psi4Mol, MethodName, BasisSet, ReturnWaveFunction = True, Quiet = OptionsInfo["QuietMode"])
    
    if not Status:
        PerformPsi4Cleanup(Psi4Handle)
        return (False, None)

    # Update atom positions...
    AtomPositions = Psi4Util.GetAtomPositions(Psi4Handle, WaveFunction, InAngstroms = True)
    RDKitUtil.SetAtomPositions(Mol, AtomPositions, ConfID = ConfID)

    # Convert energy units...
    if OptionsInfo["ApplyEnergyConversionFactor"]:
        Energy = Energy * OptionsInfo["EnergyConversionFactor"]
    
    # Clean up
    PerformPsi4Cleanup(Psi4Handle)
    
    return (True, Energy)

def MinimizeMoleculeUsingForceField(Mol, MolNum, ConfID = -1):
    """Minimize molecule using forcefield available in RDKit."""
    
    try:
        if OptionsInfo["ConfGenerationParams"]["UseUFF"]:
            ConvergeStatus = AllChem.UFFOptimizeMolecule(Mol, confId = ConfID, maxIters = OptionsInfo["ConfGenerationParams"]["MaxIters"])
        elif OptionsInfo["ConfGenerationParams"]["UseMMFF"]:
            ConvergeStatus = AllChem.MMFFOptimizeMolecule(Mol, confId = ConfID, maxIters = OptionsInfo["ConfGenerationParams"]["MaxIters"], mmffVariant = OptionsInfo["ConfGenerationParams"]["MMFFVariant"])
        else:
            MiscUtil.PrintError("Minimization couldn't be performed: Specified forcefield, %s, is not supported" % OptionsInfo["ConfGenerationParams"]["ForceField"])
    except (ValueError, RuntimeError, Chem.rdchem.KekulizeException) as ErrMsg:
        if not OptionsInfo["QuietMode"]:
            MolName = RDKitUtil.GetMolName(Mol, MolNum)
            MiscUtil.PrintWarning("Minimization using forcefield couldn't be performed for molecule %s:\n%s\n" % (MolName, ErrMsg))
        return (False, None)
    
    return (True, ConvergeStatus)

def EmbedMolecule(Mol, MolNum = None):
    """Embed conformations"""

    ConfIDs = []
    
    MaxConfs = OptionsInfo["ConfGenerationParams"]["MaxConfs"]
    RandomSeed = OptionsInfo["ConfGenerationParams"]["RandomSeed"]
    EnforceChirality = OptionsInfo["ConfGenerationParams"]["EnforceChirality"]
    UseExpTorsionAnglePrefs = OptionsInfo["ConfGenerationParams"]["UseExpTorsionAnglePrefs"]
    ETVersion = OptionsInfo["ConfGenerationParams"]["ETVersion"]
    UseBasicKnowledge = OptionsInfo["ConfGenerationParams"]["UseBasicKnowledge"]
    EmbedRMSDCutoff = OptionsInfo["ConfGenerationParams"]["EmbedRMSDCutoff"]

    try:
        ConfIDs = AllChem.EmbedMultipleConfs(Mol, numConfs = MaxConfs, randomSeed = RandomSeed, pruneRmsThresh = EmbedRMSDCutoff, enforceChirality = EnforceChirality, useExpTorsionAnglePrefs = UseExpTorsionAnglePrefs, useBasicKnowledge = UseBasicKnowledge, ETversion = ETVersion)
    except ValueError as ErrMsg:
        if not OptionsInfo["QuietMode"]:
            MolName = RDKitUtil.GetMolName(Mol, MolNum)
            MiscUtil.PrintWarning("Embedding failed  for molecule %s:\n%s\n" % (MolName, ErrMsg))
        ConfIDs = []

    if not OptionsInfo["QuietMode"]:
        if EmbedRMSDCutoff > 0:
            MiscUtil.PrintInfo("Generating initial conformation ensemble by distance geometry for %s - EmbedRMSDCutoff: %s; Size: %s; Size after RMSD filtering: %s" % (RDKitUtil.GetMolName(Mol, MolNum), EmbedRMSDCutoff, MaxConfs, len(ConfIDs)))
        else:
            MiscUtil.PrintInfo("Generating initial conformation ensemble by distance geometry for %s - EmbedRMSDCutoff: None; Size: %s" % (RDKitUtil.GetMolName(Mol, MolNum), len(ConfIDs)))
    
    return ConfIDs

def SetupPsi4Mol(Psi4Handle, Mol, MolNum, ConfID = -1):
    """Setup a Psi4 molecule object."""

    if OptionsInfo["RecenterAndReorient"]:
        MolGeometry = RDKitUtil.GetPsi4XYZFormatString(Mol, ConfID = ConfID, NoCom = False, NoReorient = False)
    else:
        MolGeometry = RDKitUtil.GetPsi4XYZFormatString(Mol, ConfID = ConfID, NoCom = True, NoReorient = True)
    
    try:
        Psi4Mol = Psi4Handle.geometry(MolGeometry)
    except Exception as ErrMsg:
        Psi4Mol = None
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintWarning("Failed to create Psi4 molecule from geometry string: %s\n" % ErrMsg)
            MolName = RDKitUtil.GetMolName(Mol, MolNum)
            MiscUtil.PrintWarning("Ignoring molecule: %s" % MolName)
    
    if OptionsInfo["Symmetrize"]:
        Psi4Mol.symmetrize(OptionsInfo["SymmetrizeTolerance"])
    
    return Psi4Mol

def PerformPsi4Cleanup(Psi4Handle):
    """Perform clean up."""

    # Clean up after Psi4 run...
    Psi4Handle.core.clean()
    
    # Clean up any leftover scratch files...
    if OptionsInfo["MPMode"]:
        Psi4Util.RemoveScratchFiles(Psi4Handle, OptionsInfo["Psi4RunParams"]["OutputFile"])

def CheckAndValidateMolecule(Mol, MolCount = None):
    """Validate molecule for Psi4 calculations."""
    
    if Mol is None:
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintInfo("\nProcessing molecule number %s..." % MolCount)
        return False
    
    MolName = RDKitUtil.GetMolName(Mol, MolCount)
    if not OptionsInfo["QuietMode"]:
        MiscUtil.PrintInfo("\nProcessing molecule %s..." % MolName)
    
    if RDKitUtil.IsMolEmpty(Mol):
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintWarning("Ignoring empty molecule: %s\n" % MolName)
        return False
    
    if not RDKitUtil.ValidateElementSymbols(RDKitUtil.GetAtomSymbols(Mol)):
        if not OptionsInfo["QuietMode"]:
            MiscUtil.PrintWarning("Ignoring molecule containing invalid element symbols: %s\n" % MolName)
        return False
    
    return True

def SetupMethodNameAndBasisSet(Mol):
    """Setup method name and basis set."""
    
    MethodName = OptionsInfo["MethodName"]
    if OptionsInfo["MethodNameAuto"]:
        MethodName = "B3LYP"
    
    BasisSet = OptionsInfo["BasisSet"]
    if OptionsInfo["BasisSetAuto"]:
        BasisSet = "6-31+G**" if RDKitUtil.IsAtomSymbolPresentInMol(Mol, "S") else "6-31G**"
    
    return (MethodName, BasisSet)

def SetupReferenceWavefunction(Mol):
    """Setup reference wavefunction."""
    
    Reference = OptionsInfo["Reference"]
    if OptionsInfo["ReferenceAuto"]:
        Reference = 'UHF' if (RDKitUtil.GetSpinMultiplicity(Mol) > 1) else 'RHF'
    
    return Reference

def SetupEnergyConversionFactor(Psi4Handle):
    """Setup converstion factor for energt units. The Psi4 energy units are Hartrees."""
    
    EnergyUnits = OptionsInfo["EnergyUnits"]
    
    ApplyConversionFactor = True
    if re.match("^kcal\/mol$", EnergyUnits, re.I):
        ConversionFactor = Psi4Handle.constants.hartree2kcalmol
    elif re.match("^kJ\/mol$", EnergyUnits, re.I):
        ConversionFactor = Psi4Handle.constants.hartree2kJmol
    elif re.match("^eV$", EnergyUnits, re.I):
        ConversionFactor = Psi4Handle.constants.hartree2ev
    else:
        ApplyConversionFactor = False
        ConversionFactor = 1.0
    
    OptionsInfo["ApplyEnergyConversionFactor"] = ApplyConversionFactor
    OptionsInfo["EnergyConversionFactor"] = ConversionFactor

def WriteMolConformers(Writer, Mol, MolNum, ConfIDs, ConfEnergies = None):
    """Write molecule coformers."""

    if ConfIDs is None:
        return True
    
    MolName = RDKitUtil.GetMolName(Mol, MolNum)

    for Index, ConfID in enumerate(ConfIDs):
        SetConfMolName(Mol, MolName, ConfID)
        
        if OptionsInfo["EnergyOut"]  and ConfEnergies is not None:
            Mol.SetProp(OptionsInfo["EnergyDataFieldLabel"], ConfEnergies[Index])
            
            Writer.write(Mol, confId = ConfID)

def SetConfMolName(Mol, MolName, ConfCount):
    """Set conf mol name."""

    ConfName = "%s_Conf%d" % (MolName, ConfCount)
    Mol.SetProp("_Name", ConfName)


def ProcessOptions():
    """Process and validate command line arguments and options."""
    
    MiscUtil.PrintInfo("Processing options...")
    
    # Validate options...
    ValidateOptions()
    
    OptionsInfo["Infile"] = Options["--infile"]
    OptionsInfo["SMILESInfileStatus"] = True if  MiscUtil.CheckFileExt(Options["--infile"], "smi csv tsv txt") else False
    ParamsDefaultInfoOverride = {"RemoveHydrogens": False}
    OptionsInfo["InfileParams"] = MiscUtil.ProcessOptionInfileParameters("--infileParams", Options["--infileParams"], InfileName = Options["--infile"], ParamsDefaultInfo = ParamsDefaultInfoOverride)
    
    OptionsInfo["Outfile"] = Options["--outfile"]
    OptionsInfo["OutfileParams"] = MiscUtil.ProcessOptionOutfileParameters("--outfileParams", Options["--outfileParams"])
    
    OptionsInfo["Overwrite"] = Options["--overwrite"]
    
    # Method, basis set, and reference wavefunction...
    OptionsInfo["BasisSet"] = Options["--basisSet"]
    OptionsInfo["BasisSetAuto"] = True if re.match("^auto$", Options["--basisSet"], re.I) else False
    
    OptionsInfo["MethodName"] = Options["--methodName"]
    OptionsInfo["MethodNameAuto"] = True if re.match("^auto$", Options["--methodName"], re.I) else False
    
    OptionsInfo["Reference"] = Options["--reference"]
    OptionsInfo["ReferenceAuto"] = True if re.match("^auto$", Options["--reference"], re.I) else False
    
    # Run and options parameters...
    OptionsInfo["Psi4OptionsParams"] = Psi4Util.ProcessPsi4OptionsParameters("--psi4OptionsParams", Options["--psi4OptionsParams"])
    OptionsInfo["Psi4RunParams"] = Psi4Util.ProcessPsi4RunParameters("--psi4RunParams", Options["--psi4RunParams"], InfileName = OptionsInfo["Infile"])

    # Conformer generation paramaters...
    ParamsDefaultInfoOverride = {"MaxConfs": 50, "MaxIters": 250}
    OptionsInfo["ConfGenerationParams"] = MiscUtil.ProcessOptionConformerParameters("--confParams", Options["--confParams"], ParamsDefaultInfoOverride)
    
    # Energy parameters...
    OptionsInfo["EnergyOut"] = True if re.match("^yes$", Options["--energyOut"], re.I) else False
    OptionsInfo["EnergyUnits"] = Options["--energyUnits"]
    
    EnergyDataFieldLabel = Options["--energyDataFieldLabel"]
    if re.match("^auto$", EnergyDataFieldLabel, re.I):
        EnergyDataFieldLabel = "Psi4_Energy (%s)" % Options["--energyUnits"]
    OptionsInfo["EnergyDataFieldLabel"] = EnergyDataFieldLabel
    
    OptionsInfo["EnergyRMSDCalcMode"] = Options["--energyRMSDCalcMode"]
    OptionsInfo["EnergyRMSDCalcModeBest"] = True if re.match("^BestRMSD$", Options["--energyRMSDCalcMode"], re.I) else False
    
    OptionsInfo["EnergyRMSDCutoff"] = float(Options["--energyRMSDCutoff"])
    
    OptionsInfo["EnergyRMSDCutoffMode"] = Options["--energyRMSDCutoffMode"]
    OptionsInfo["EnergyRMSDCutoffModeLowest"] = True if re.match("^Lowest$", Options["--energyRMSDCutoffMode"], re.I) else False
    
    if OptionsInfo["EnergyRMSDCutoff"] > 0:
        # Make sure that the alignConformers option is being used...
        if not OptionsInfo["ConfGenerationParams"]["AlignConformers"]:
            MiscUtil.PrintError("The value for \"alignConformers\" specified using \"--confParams\" must  be set to \" yes\" for non-zero values of \"--energyRMSDCutoff\" ")
    
    # Process energy window...
    EnergyWindow = Options["--energyWindow"]
    if re.match("^auto$", EnergyWindow, re.I):
        # Set default energy window based on units...
        EnergyUnits = Options["--energyUnits"]
        if re.match("^kcal\/mol$", EnergyUnits, re.I):
            EnergyWindow = 20
        elif re.match("^kJ\/mol$", EnergyUnits, re.I):
            EnergyWindow = 83.68
        elif re.match("^eV$", EnergyUnits, re.I):
            EnergyWindow = 0.8673
        elif re.match("^Hartrees$", EnergyUnits, re.I):
            EnergyWindow = 0.03188
        else:
            MiscUtil.PrintError("Failed to set default value for \"--energyWindow\". The value, %s, specified for option \"--energyUnits\" is not valid. " % EnergyUnits)
    else:
        EnergyWindow = float(EnergyWindow)
    OptionsInfo["EnergyWindow"] = EnergyWindow
    
    OptionsInfo["MaxIters"] = int(Options["--maxIters"])

    OptionsInfo["MPMode"] = True if re.match("^yes$", Options["--mp"], re.I) else False
    OptionsInfo["MPParams"] = MiscUtil.ProcessOptionMultiprocessingParameters("--mpParams", Options["--mpParams"])

    # Multiprocessing level...
    MPLevelMoleculesMode = False
    MPLevelConformersMode = False
    MPLevel = Options["--mpLevel"]
    if re.match("^Molecules$", MPLevel, re.I):
        MPLevelMoleculesMode = True
    elif re.match("^Conformers$", MPLevel, re.I):
        MPLevelConformersMode = True
    else:
        MiscUtil.PrintError("The value, %s, specified for option \"--mpLevel\" is not valid. " % MPLevel)
    OptionsInfo["MPLevel"] = MPLevel
    OptionsInfo["MPLevelMoleculesMode"] = MPLevelMoleculesMode
    OptionsInfo["MPLevelConformersMode"] = MPLevelConformersMode
    
    OptionsInfo["Precision"] = int(Options["--precision"])
    OptionsInfo["QuietMode"] = True if re.match("^yes$", Options["--quiet"], re.I) else False

    OptionsInfo["RecenterAndReorient"] = True if re.match("^yes$", Options["--recenterAndReorient"], re.I) else False
    
    Symmetrize = Options["--symmetrize"]
    if re.match("^auto$", Symmetrize, re.I):
        Symmetrize  = True if  OptionsInfo["RecenterAndReorient"] else False
    else:
        Symmetrize  = True if re.match("^yes$", Symmetrize, re.I) else False
    OptionsInfo["Symmetrize"] = Symmetrize
    
    OptionsInfo["SymmetrizeTolerance"] = float(Options["--symmetrizeTolerance"])

def RetrieveOptions():
    """Retrieve command line arguments and options."""
    
    # Get options...
    global Options
    Options = docopt(_docoptUsage_)
    
    # Set current working directory to the specified directory...
    WorkingDir = Options["--workingdir"]
    if WorkingDir:
        os.chdir(WorkingDir)
    
    # Handle examples option...
    if "--examples" in Options and Options["--examples"]:
        MiscUtil.PrintInfo(MiscUtil.GetExamplesTextFromDocOptText(_docoptUsage_))
        sys.exit(0)

def ValidateOptions():
    """Validate option values."""
    
    MiscUtil.ValidateOptionTextValue("--energyOut", Options["--energyOut"], "yes no")
    MiscUtil.ValidateOptionTextValue("--energyUnits", Options["--energyUnits"], "Hartrees kcal/mol kJ/mol eV")
    
    MiscUtil.ValidateOptionTextValue(" --energyRMSDCalcMode", Options["--energyRMSDCalcMode"], "RMSD BestRMSD")
    
    MiscUtil.ValidateOptionFloatValue("--energyRMSDCutoff", Options["--energyRMSDCutoff"], {">=": 0})
    MiscUtil.ValidateOptionTextValue(" --energyRMSDCutoffMode", Options["--energyRMSDCutoffMode"], "All Lowest")
    
    if not re.match("^auto$", Options["--energyWindow"], re.I):
        MiscUtil.ValidateOptionFloatValue("--energyWindow", Options["--energyWindow"], {">": 0})
    
    MiscUtil.ValidateOptionFilePath("-i, --infile", Options["--infile"])
    MiscUtil.ValidateOptionFileExt("-i, --infile", Options["--infile"], "sdf sd mol smi txt csv tsv")
    
    MiscUtil.ValidateOptionIntegerValue("--maxIters", Options["--maxIters"], {">": 0})
    
    MiscUtil.ValidateOptionFileExt("-o, --outfile", Options["--outfile"], "sdf sd")
    MiscUtil.ValidateOptionsOutputFileOverwrite("-o, --outfile", Options["--outfile"], "--overwrite", Options["--overwrite"])
    MiscUtil.ValidateOptionsDistinctFileNames("-i, --infile", Options["--infile"], "-o, --outfile", Options["--outfile"])
    
    MiscUtil.ValidateOptionTextValue("--mp", Options["--mp"], "yes no")
    MiscUtil.ValidateOptionTextValue("--mpLevel", Options["--mpLevel"], "Molecules Conformers")
    
    MiscUtil.ValidateOptionIntegerValue("-p, --precision", Options["--precision"], {">": 0})
    MiscUtil.ValidateOptionTextValue("-q, --quiet", Options["--quiet"], "yes no")
    
    MiscUtil.ValidateOptionTextValue("--recenterAndReorient", Options["--recenterAndReorient"], "yes no")
    MiscUtil.ValidateOptionTextValue("--symmetrize", Options["--symmetrize"], "yes no auto")
    MiscUtil.ValidateOptionFloatValue("--symmetrizeTolerance", Options["--symmetrizeTolerance"], {">": 0})

# Setup a usage string for docopt...
_docoptUsage_ = """
Psi4GenerateConformers.py - Generate molecular conformations

Usage:
    Psi4GenerateConformers.py [--basisSet <text>] [--confParams <Name,Value,...>] [--energyOut <yes or no>]
                              [--energyDataFieldLabel <text>] [--energyUnits <text>] [--energyRMSDCalcMode <RMSD or BestRMSD>]
                              [--energyRMSDCutoff <number>] [--energyRMSDCutoffMode <All or Lowest>] [--energyWindow <number>]
                              [--infileParams <Name,Value,...>] [--maxIters <number>]
                              [--methodName <text>] [--mp <yes or no>] [--mpLevel <Molecules or Conformers>]
                              [--mpParams <Name, Value,...>] [ --outfileParams <Name,Value,...> ] [--overwrite] [--precision <number>]
                              [--psi4OptionsParams <Name,Value,...>] [--psi4RunParams <Name,Value,...>]
                              [--quiet <yes or no>] [--reference <text>] [--recenterAndReorient <yes or no>]
                              [--symmetrize <yes or no>] [--symmetrizeTolerance <number>] [-w <dir>] -i <infile> -o <outfile> 
    Psi4GenerateConformers.py -h | --help | -e | --examples

Description:
    Generate 3D conformers of molecules using a combination of distance geometry
    and forcefield minimization followed by geometry optimization using a quantum
    chemistry method. A set of initial 3D structures are generated for a molecule 
    employing distance geometry. The 3D structures in the conformation ensemble
    are sequentially minimized using forcefield and a quantum chemistry method.

    A Psi4 XYZ format geometry string is automatically generated for each molecule
    in input file. It contains atom symbols and 3D coordinates for each atom in a
    molecule. In addition, the formal charge and spin multiplicity are present in the
    the geometry string. These values are either retrieved from molecule properties
    named 'FormalCharge' and 'SpinMultiplicty' or dynamically calculated for a
    molecule.

    The supported input file formats are: Mol (.mol), SD (.sdf, .sd), SMILES (.smi,
    .csv, .tsv .txt)

    The supported output file formats are: SD (.sdf, .sd)

Options:
    -b, --basisSet <text>  [default: auto]
        Basis set to use for energy minimization. Default: 6-31+G** for sulfur
        containing molecules; Otherwise, 6-31G** [ Ref 150 ]. The specified 
        value must be a valid Psi4 basis set. No validation is performed.
        
        The following list shows a representative sample of basis sets available
        in Psi4:
            
            STO-3G, 6-31G, 6-31+G, 6-31++G, 6-31G*, 6-31+G*,  6-31++G*, 
            6-31G**, 6-31+G**, 6-31++G**, 6-311G, 6-311+G, 6-311++G,
            6-311G*, 6-311+G*, 6-311++G*, 6-311G**, 6-311+G**, 6-311++G**,
            cc-pVDZ, cc-pCVDZ, aug-cc-pVDZ, cc-pVDZ-DK, cc-pCVDZ-DK, def2-SVP,
            def2-SVPD, def2-TZVP, def2-TZVPD, def2-TZVPP, def2-TZVPPD
            
    --confParams <Name,Value,...>  [default: auto]
        Generate an initial 3D conformation ensemble using distance geometry and
        forcefield minimization before final geometry optimization by a specified
        method name and basis set. Possible values: yes or no.
        
        A comma delimited list of parameter name and value pairs for generating
        initial sets of 3D conformations for molecules. The 3D conformation ensemble
        is generated using distance geometry and forcefield functionality available
        in RDKit. The 3D structures in the conformation ensemble are subsequently
        minimized by a quantum chemistry method available in Psi4.
       
        The supported parameter names along with their default values are shown
        below:
            
            confMethod,ETKDGv2,
            forceField,MMFF, forceFieldMMFFVariant,MMFF94,
            enforceChirality,yes,alignConformers,yes, embedRMSDCutoff,0.5,
            maxConfs,50,maxIters,250,randomSeed,auto
            
            confMethod,ETKDGv2   [ Possible values: SDG, KDG, ETDG,
                ETKDG , or ETKDGv2]
            forceField, MMFF   [ Possible values: UFF or MMFF ]
            forceFieldMMFFVariant,MMFF94   [ Possible values: MMFF94 or MMFF94s ]
            enforceChirality,yes   [ Possible values: yes or no ]
            alignConformers,yes   [ Possible values: yes or no ]
            embedRMSDCutoff,0.5   [ Possible values: number or None]
            
        confMethod: Conformation generation methodology for generating initial 3D
        coordinates. Possible values: Standard Distance Geometry (SDG), Experimental
        Torsion-angle preference with Distance Geometry (ETDG), basic Knowledge-terms
        with Distance Geometry (KDG) and Experimental Torsion-angle preference
        along with basic Knowledge-terms with Distance Geometry (ETKDG or
        ETKDGv2) [Ref 129, 167] .
        
        forceField: Forcefield method to use for energy minimization. Possible
        values: Universal Force Field (UFF) [ Ref 81 ] or Merck Molecular Mechanics
        Force Field [ Ref 83-87 ] .
        
        enforceChirality: Enforce chirality for defined chiral centers during
        forcefield minimization.
        
        alignConformers: Align conformers for each molecule.
        
        maxConfs: Maximum number of conformations to generate for each molecule
        during the generation of an initial 3D conformation ensemble using 
        conformation generation methodology. The conformations are minimized
        using the specified forcefield and a quantum chemistry method. The lowest
        energy conformation is written to the output file.
        
        embedRMSDCutoff: RMSD cutoff for retaining initial set of conformers embedded
        using distance geometry and before forcefield minimization. All embedded
        conformers are kept for 'None' value. Otherwise, only those conformers which
        are different from each other by the specified RMSD cutoff, 0.5 by default,
        are kept. The first embedded conformer is always retained.
        
        maxIters: Maximum number of iterations to perform for each conformation
        during forcefield minimization.
        
        randomSeed: Seed for the random number generator for reproducing initial
        3D coordinates in a conformation ensemble. Default is to use a random seed.
    --energyOut <yes or no>  [default: yes]
        Write out energy values.
    --energyDataFieldLabel <text>  [default: auto]
        Energy data field label for writing energy values. Default: Psi4_Energy (<Units>). 
    --energyUnits <text>  [default: kcal/mol]
        Energy units. Possible values: Hartrees, kcal/mol, kJ/mol, or eV.
    --energyRMSDCalcMode <RMSD or BestRMSD>  [default: RMSD]
        Methodology for calculating RMSD values during the application of RMSD
        cutoff for retaining conformations after the final energy minimization. Possible
        values: RMSD or BestRMSD. This option is ignore during 'None' value of
        '--energyRMSDCutoff' option.
        
        During BestRMSMode mode, the RDKit 'function AllChem.GetBestRMS' is used to
        align and calculate RMSD. This function calculates optimal RMSD for aligning two
        molecules, taking symmetry into account. Otherwise, the RMSD value is calculated
        using 'AllChem.GetConformerRMS' without changing the atom order. A word to the
        wise from RDKit documentation: The AllChem.GetBestRMS function will attempt to
        align all permutations of matching atom orders in both molecules, for some molecules
        it will lead to 'combinatorial explosion'.
    --energyRMSDCutoff <number>  [default: 0.5]
        RMSD cutoff for retaining conformations after the final energy minimization.
        By default, only those conformations which are different from the lowest
        energy conformation by the specified RMSD cutoff and are with in the 
        specified energy window are kept. The lowest energy conformation is always
        retained. A value of zero keeps all minimized conformations with in the
        specified energy window from the lowest energy.
    --energyRMSDCutoffMode <All or Lowest>  [default: All]
        RMSD cutoff mode for  retaining conformations after the final energy
        minimization. Possible values: All or Lowest. The RMSD values are compared
        against all the selected conformations or the lowest energy conformation during
        'All' and 'Lowest' value of '--energyRMSDCutoffMode'. This option is ignored
        during 'None' value of --energyRMSDCutoff.
        
        By default, only those conformations which all different from all selected
        conformations by the specified RMSD cutoff and are with in the specified
        energy window are kept.
    --energyWindow <number>  [default: auto]
        Psi4 Energy window  for selecting conformers after the final energy minimization.
        The default value is dependent on '--energyUnits': 20 kcal/mol, 83.68 kJ/mol,
        0.8673 ev, or 0.03188 Hartrees. The specified value must be in '--energyUnits'.
    -e, --examples
        Print examples.
    -h, --help
        Print this help message.
    -i, --infile <infile>
        Input file name.
    --infileParams <Name,Value,...>  [default: auto]
        A comma delimited list of parameter name and value pairs for reading
        molecules from files. The supported parameter names for different file
        formats, along with their default values, are shown below:
            
            SD, MOL: removeHydrogens,no,sanitize,yes,strictParsing,yes
            SMILES: smilesColumn,1,smilesNameColumn,2,smilesDelimiter,space,
                smilesTitleLine,auto,sanitize,yes
            
        Possible values for smilesDelimiter: space, comma or tab.
    --maxIters <number>  [default: 50]
        Maximum number of iterations to perform for each molecule or conformer
        during energy minimization by a quantum chemistry method.
    -m, --methodName <text>  [default: auto]
        Method to use for energy minimization. Default: B3LYP [ Ref 150 ]. The
        specified value must be a valid Psi4 method name. No validation is
        performed.
        
        The following list shows a representative sample of methods available
        in Psi4:
            
            B1LYP, B2PLYP, B2PLYP-D3BJ, B2PLYP-D3MBJ, B3LYP, B3LYP-D3BJ,
            B3LYP-D3MBJ, CAM-B3LYP, CAM-B3LYP-D3BJ, HF, HF-D3BJ,  HF3c, M05,
            M06, M06-2x, M06-HF, M06-L, MN12-L, MN15, MN15-D3BJ,PBE, PBE0,
            PBEH3c, PW6B95, PW6B95-D3BJ, WB97, WB97X, WB97X-D, WB97X-D3BJ
            
    --mp <yes or no>  [default: no]
        Use multiprocessing.
         
        By default, input data is retrieved in a lazy manner via mp.Pool.imap()
        function employing lazy RDKit data iterable. This allows processing of
        arbitrary large data sets without any additional requirements memory.
        
        All input data may be optionally loaded into memory by mp.Pool.map()
        before starting worker processes in a process pool by setting the value
        of 'inputDataMode' to 'InMemory' in '--mpParams' option.
        
        A word to the wise: The default 'chunkSize' value of 1 during 'Lazy' input
        data mode may adversely impact the performance. The '--mpParams' section
        provides additional information to tune the value of 'chunkSize'.
    --mpLevel <Molecules or Conformers>  [default: Molecules]
        Perform multiprocessing at molecules or conformers level. Possible values:
        Molecules or Conformers. The 'Molecules' value starts a process pool at the
        molecules level. All conformers of a molecule are processed in a single
        process. The 'Conformers' value, however, starts a process pool at the 
        conformers level. Each conformer of a molecule is processed in an individual
        process in the process pool. The default Psi4 'OutputFile' is set to 'quiet'
        using '--psi4RunParams' for 'Conformers' level. Otherwise, it may generate
        a large number of Psi4 output files.
    --mpParams <Name,Value,...>  [default: auto]
        A comma delimited list of parameter name and value pairs to configure
        multiprocessing.
        
        The supported parameter names along with their default and possible
        values are shown below:
        
            chunkSize, auto
            inputDataMode, Lazy   [ Possible values: InMemory or Lazy ]
            numProcesses, auto   [ Default: mp.cpu_count() ]
        
        These parameters are used by the following functions to configure and
        control the behavior of multiprocessing: mp.Pool(), mp.Pool.map(), and
        mp.Pool.imap().
        
        The chunkSize determines chunks of input data passed to each worker
        process in a process pool by mp.Pool.map() and mp.Pool.imap() functions.
        The default value of chunkSize is dependent on the value of 'inputDataMode'.
        
        The mp.Pool.map() function, invoked during 'InMemory' input data mode,
        automatically converts RDKit data iterable into a list, loads all data into
        memory, and calculates the default chunkSize using the following method
        as shown in its code:
        
            chunkSize, extra = divmod(len(dataIterable), len(numProcesses) * 4)
            if extra: chunkSize += 1
        
        For example, the default chunkSize will be 7 for a pool of 4 worker processes
        and 100 data items.
        
        The mp.Pool.imap() function, invoked during 'Lazy' input data mode, employs
        'lazy' RDKit data iterable to retrieve data as needed, without loading all the
        data into memory. Consequently, the size of input data is not known a priori.
        It's not possible to estimate an optimal value for the chunkSize. The default 
        chunkSize is set to 1.
        
        The default value for the chunkSize during 'Lazy' data mode may adversely
        impact the performance due to the overhead associated with exchanging
        small chunks of data. It is generally a good idea to explicitly set chunkSize to
        a larger value during 'Lazy' input data mode, based on the size of your input
        data and number of processes in the process pool.
        
        The mp.Pool.map() function waits for all worker processes to process all
        the data and return the results. The mp.Pool.imap() function, however,
        returns the the results obtained from worker processes as soon as the
        results become available for specified chunks of data.
        
        The order of data in the results returned by both mp.Pool.map() and 
        mp.Pool.imap() functions always corresponds to the input data.
    -o, --outfile <outfile>
        Output file name.
    --outfileParams <Name,Value,...>  [default: auto]
        A comma delimited list of parameter name and value pairs for writing
        molecules to files. The supported parameter names for different file
        formats, along with their default values, are shown below:
            
            SD: kekulize,yes,forceV3000,no
            
    --overwrite
        Overwrite existing files.
    --precision <number>  [default: 6]
        Floating point precision for writing energy values.
    --psi4OptionsParams <Name,Value,...>  [default: none]
        A comma delimited list of Psi4 option name and value pairs for setting
        global and module options. The names are 'option_name' for global options
        and 'module_name__option_name' for options local to a module. The
        specified option names must be valid Psi4 names. No validation is
        performed.
        
        The specified option name and  value pairs are processed and passed to
        psi4.set_options() as a dictionary. The supported value types are float,
        integer, boolean, or string. The float value string is converted into a float.
        The valid values for a boolean string are yes, no, true, false, on, or off. 
    --psi4RunParams <Name,Value,...>  [default: auto]
        A comma delimited list of parameter name and value pairs for configuring
        Psi4 jobs.
        
        The supported parameter names along with their default and possible
        values are shown below:
             
            MemoryInGB, 1
            NumThreads, 1
            OutputFile, auto   [ Possible  values: stdout, quiet, or FileName ]
            ScratchDir, auto   [ Possivle values: DirName]
            RemoveOutputFile, yes   [ Possible values: yes, no, true, or false]
            
        These parameters control the runtime behavior of Psi4.
        
        The default file name for 'OutputFile' is <InFileRoot>_Psi4.out. The PID
        is appended to output file name during multiprocessing as shown:
        <InFileRoot>_Psi4_<PIDNum>.out. The 'stdout' value for 'OutputType'
        sends Psi4 output to stdout. The 'quiet' or 'devnull' value suppresses
        all Psi4 output. The 'OutputFile' is set to 'quiet' for 'auto' value during 
        'Conformers' of '--mpLevel' option.
        
        The default 'Yes' value of 'RemoveOutputFile' option forces the removal
        of any existing Psi4 before creating new files to append output from
        multiple Psi4 runs.
        
        The option 'ScratchDir' is a directory path to the location of scratch
        files. The default value corresponds to Psi4 default. It may be used to
        override the deafult path.
    -q, --quiet <yes or no>  [default: no]
        Use quiet mode. The warning and information messages will not be printed.
    -r, --reference <text>  [default: auto]
        Reference wave function to use for energy calculation. Default: RHF or UHF.
        The default values are Restricted Hartree-Fock (RHF) for closed-shell molecules
        with all electrons paired and Unrestricted Hartree-Fock (UHF) for open-shell
        molecules with unpaired electrons.
        
        The specified value must be a valid Psi4 reference wave function. No validation
        is performed. For example: ROHF, CUHF, RKS, etc.
        
        The spin multiplicity determines the default value of reference wave function
        for input molecules. It is calculated from number of free radical electrons using
        Hund's rule of maximum multiplicity defined as 2S + 1 where S is the total
        electron spin. The total spin is 1/2 the number of free radical electrons in a 
        molecule. The value of 'SpinMultiplicity' molecule property takes precedence
        over the calculated value of spin multiplicity.
    --recenterAndReorient <yes or no>  [default: yes]
        Recenter and reorient a molecule during creation of a Psi4 molecule from
        a geometry string.
        
        The 'No' values allows the minimization of a molecule in its initial 3D
        coordinate space generated by RDKit.
    --symmetrize <yes or no>  [default: auto]
        Symmetrize molecules before energy minimization. Default: 'Yes' during
        'Yes' value of '--recenterAndReorient'; Otherwise, 'No'. The psi4 function,
        psi4mol.symmetrize( SymmetrizeTolerance), is called to symmetrize
        the molecule before calling psi4.optimize().
        
        The 'No' value of '--symmetrize' during 'Yes' value of '--recenterAndReorient'
        may cause psi4.optimize() to fail with a 'Point group changed...' error
        message.
    --symmetrizeTolerance <number>  [default: 0.01]
        Symmetry tolerance for '--symmetrize'.
    -w, --workingdir <dir>
        Location of working directory which defaults to the current directory.

Examples:
    To generate an initial conformer ensemble of up to 50 conformations using a
    combination of ETKDGv2 distance geometry methodology, applying embed RMSD
    cutoff of 0.5 and MMFF forcefield minimization, followed by energy minimization
    using B3LYP/6-31G** or B3LYP/6-31+G** (sulfur containing), selecting a final set
    of minimized conformers for molecules in a SMILES file, applying energy RMSD
    cutoff of 0.5 and energy window value value of 20 kcal/mol, and write out a SD
    file containing minimized conformers, type:

        % Psi4GenerateConformers.py -i Psi4Sample.smi -o Psi4SampleOut.sdf

    To run the first example in a quiet mode and write out a SD file, type:

        % Psi4GenerateConformers.py -q yes -i Psi4Sample.smi -o
          Psi4SampleOut.sdf

    To run the first example in multiprocessing mode at molecules level on all
    available CPUs without loading all data into memory and write out a SD file,
    type:

        % Psi4GenerateConformers.py --mp yes -i Psi4Sample.smi -o
          Psi4SampleOut.sdf

    To run the first example in multiprocessing mode at conformers level on all
    available CPUs without loading all data into memory and write out a SD file,
    type:

        % Psi4GenerateConformers.py --mp yes --mpLevel Conformers
          -i Psi4Sample.smi -o Psi4SampleOut.sdf

    To run the first example in multiprocessing mode at molecules level on specific
    number of CPUs and chunk size without loading all data into memory and write
    out a SD file, type:

        % Psi4GenerateConformers.py  --mp yes --mpParams "inputDataMode,Lazy,
          numProcesses,4,chunkSize,8" -i Psi4Sample.smi -o Psi4SampleOut.sdf

    To run the first example by using an explicit set of specific parameters, and
    write out a SD file, type

        % Psi4GenerateConformers.py --confParams "confMethod,ETKDGv2,
          forceField,MMFF, forceFieldMMFFVariant,MMFF94s, maxConfs,20,
          embedRMSDCutoff,0.25" --energyUnits "kJ/mol" -m B3LYP
          -b "6-31+G**" --maxIters 20 -i Psi4Sample.smi -o Psi4SampleOut.sdf

    To run the first example for molecules in a CSV SMILES file, SMILES strings
    in column 1, name column 2, and write out a SD file, type:

        % Psi4GenerateConformers.py --infileParams "smilesDelimiter,comma,
          smilesTitleLine,yes,smilesColumn,1,smilesNameColumn,2"
          -i Psi4Sample.csv -o Psi4SampleOut.sdf

Author:

    Manish Sud(msud@san.rr.com)

See also:
    Psi4CalculateEnergy.py, Psi4CalculatePartialCharges.py, Psi4PerformMinimization.py

Copyright:
    Copyright (C) 2024 Manish Sud. All rights reserved.

    The functionality available in this script is implemented using Psi4, an
    open source quantum chemistry software package, and RDKit, an open
    source toolkit for cheminformatics developed by Greg Landrum.

    This file is part of MayaChemTools.

    MayaChemTools is free software; you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License as published by the Free
    Software Foundation; either version 3 of the License, or (at your option) any
    later version.

"""

if __name__ == "__main__":
    main()
